# Copyright 2015-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
from builtins import next
from builtins import range
import numpy as np
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.baseclasses import MainHierarchicalParser
import nomadcore.configurationreading
import nomadcore.csvparsing
from .commonparser import CP2KCommonParser
from nomadcore.caching_backend import CachingLevel
from nomadcore.unit_conversion.unit_conversion import convert_unit
import logging
logger = logging.getLogger("nomad")


class CP2KMDParser(MainHierarchicalParser):
    """Used to parse the CP2K calculation with run types:
        -MD
        -MOLECULAR_DYNAMICS
    """
    def __init__(self, parser_context):
        """
        """
        super(CP2KMDParser, self).__init__(parser_context)
        self.setup_common_matcher(CP2KCommonParser(parser_context))
        self.traj_iterator = None
        self.vel_iterator = None
        self.energy_iterator = None
        self.cell_iterator = None
        self.n_steps = None
        self.output_freq = None
        self.coord_freq = None
        self.velo_freq = None
        self.energy_freq = None
        self.cell_freq = None
        self.md_quicksteps = []
        self.ensemble = None
        self.cp2k_ensemble = None

        #=======================================================================
        # Globally cached values

        #=======================================================================
        # Cache levels
        self.caching_levels.update({
            'x_cp2k_section_md_settings': CachingLevel.ForwardAndCache,
            'x_cp2k_section_md_step': CachingLevel.ForwardAndCache,
            'x_cp2k_section_quickstep_calculation': CachingLevel.ForwardAndCache,
            "x_cp2k_section_scf_iteration": CachingLevel.ForwardAndCache,
            "x_cp2k_section_stress_tensor": CachingLevel.ForwardAndCache,
        })

        #=======================================================================
        # SimpleMatchers
        self.md = SM(
            " MD\| Molecular Dynamics Protocol",
            forwardMatch=True,
            sections=["section_frame_sequence", "x_cp2k_section_md"],
            subMatchers=[
                SM( " MD\| Molecular Dynamics Protocol",
                    forwardMatch=True,
                    sections=["section_sampling_method", "x_cp2k_section_md_settings"],
                    subMatchers=[
                        SM( " MD\| Ensemble Type\s+(?P<x_cp2k_md_ensemble_type>{})".format(self.regexs.word)),
                        SM( " MD\| Number of Time Steps\s+(?P<x_cp2k_md_number_of_time_steps>{})".format(self.regexs.int)),
                        SM( " MD\| Time Step \[fs\]\s+(?P<x_cp2k_md_time_step__fs>{})".format(self.regexs.float)),
                        SM( " MD\| Temperature \[K\]\s+(?P<x_cp2k_md_target_temperature>{})".format(self.regexs.float)),
                        SM( " MD\| Temperature tolerance \[K\]\s+(?P<x_cp2k_md_target_temperature_tolerance>{})".format(self.regexs.float)),
                        SM( " MD\| Pressure \[Bar\]\s+(?P<x_cp2k_md_target_pressure>{})".format(self.regexs.float)),
                        SM( " MD\| Barostat time constant \[  fs\]\s+(?P<x_cp2k_md_barostat_time_constant>{})".format(self.regexs.float)),
                        SM( " MD\| Print MD information every\s+(?P<x_cp2k_md_print_frequency>{}) step\(s\)".format(self.regexs.int)),
                        SM( " MD\| File type     Print frequency\[steps\]                             File names"),
                        SM( " MD\| Coordinates\s+(?P<x_cp2k_md_coordinates_print_frequency>{})\s+(?P<x_cp2k_md_coordinates_filename>{})".format(self.regexs.int, self.regexs.word)),
                        SM( " MD\| Simulation Cel\s+(?P<x_cp2k_md_simulation_cell_print_frequency>{})\s+(?P<x_cp2k_md_simulation_cell_filename>{})".format(self.regexs.int, self.regexs.word)),
                        SM( " MD\| Velocities\s+(?P<x_cp2k_md_velocities_print_frequency>{})\s+(?P<x_cp2k_md_velocities_filename>{})".format(self.regexs.int, self.regexs.word)),
                        SM( " MD\| Energies\s+(?P<x_cp2k_md_energies_print_frequency>{})\s+(?P<x_cp2k_md_energies_filename>{})".format(self.regexs.int, self.regexs.word)),
                        SM( " MD\| Dump\s+(?P<x_cp2k_md_dump_print_frequency>{})\s+(?P<x_cp2k_md_dump_filename>{})".format(self.regexs.int, self.regexs.word)),
                    ]
                ),
                SM( " ************************** Velocities initialization **************************".replace("*", "\*"),
                    name="md_initialization_step",
                    endReStr=" INITIAL CELL ANGLS",
                    sections=["x_cp2k_section_md_step"],
                    subMatchers=[
                        self.cm.quickstep_calculation(),
                        SM( " ******************************** GO CP2K GO! **********************************".replace("*", "\*")),
                        SM( " INITIAL POTENTIAL ENERGY\[hartree\]     =\s+(?P<x_cp2k_md_potential_energy_instantaneous__hartree>{})".format(self.regexs.float)),
                        SM( " INITIAL KINETIC ENERGY\[hartree\]       =\s+(?P<x_cp2k_md_kinetic_energy_instantaneous__hartree>{})".format(self.regexs.float)),
                        SM( " INITIAL TEMPERATURE\[K\]                =\s+(?P<x_cp2k_md_temperature_instantaneous>{})".format(self.regexs.float)),
                        SM( " INITIAL BAROSTAT TEMP\[K\]              =\s+(?P<x_cp2k_md_barostat_temperature_instantaneous>{})".format(self.regexs.float)),
                        SM( " INITIAL PRESSURE\[bar\]                 =\s+(?P<x_cp2k_md_pressure_instantaneous__bar>{})".format(self.regexs.float)),
                        SM( " INITIAL VOLUME\[bohr\^3\]                =\s+(?P<x_cp2k_md_volume_instantaneous__bohr3>{})".format(self.regexs.float)),
                        SM( " INITIAL CELL LNTHS\[bohr\]              =\s+(?P<x_cp2k_md_cell_length_a_instantaneous__bohr>{0})\s+(?P<x_cp2k_md_cell_length_b_instantaneous__bohr>{0})\s+(?P<x_cp2k_md_cell_length_c_instantaneous__bohr>{0})".format(self.regexs.float)),
                        SM( " INITIAL CELL ANGLS\[deg\]               =\s+(?P<x_cp2k_md_cell_angle_a_instantaneous__deg>{0})\s+(?P<x_cp2k_md_cell_angle_b_instantaneous__deg>{0})\s+(?P<x_cp2k_md_cell_angle_c_instantaneous__deg>{0})".format(self.regexs.float)),
                    ],
                    adHoc=self.adHoc_save_md_quickstep()
                ),
                SM( " SCF WAVEFUNCTION OPTIMIZATION",
                    endReStr=" TEMPERATURE \[K\]              =",
                    name="md_step",
                    forwardMatch=True,
                    repeats=True,
                    sections=["x_cp2k_section_md_step"],
                    subMatchers=[
                        self.cm.quickstep_calculation(),
                        SM( " ENSEMBLE TYPE                ="),
                        SM( " STEP NUMBER                  =\s+(?P<x_cp2k_md_step_number>{})".format(self.regexs.int)),
                        SM( " TIME \[fs\]                    =\s+(?P<x_cp2k_md_time__fs>{})".format(self.regexs.float)),
                        SM( " CONSERVED QUANTITY \[hartree\] =\s+(?P<x_cp2k_md_conserved_quantity__hartree>{})".format(self.regexs.float)),
                        SM( " CPU TIME \[s\]                 =\s+(?P<x_cp2k_md_cpu_time_instantaneous>{})\s+(?P<x_cp2k_md_cpu_time_average>{})".format(self.regexs.float, self.regexs.float)),
                        SM( " ENERGY DRIFT PER ATOM \[K\]    =\s+(?P<x_cp2k_md_energy_drift_instantaneous>{})\s+(?P<x_cp2k_md_energy_drift_average>{})".format(self.regexs.float, self.regexs.float)),
                        SM( " POTENTIAL ENERGY\[hartree\]    =\s+(?P<x_cp2k_md_potential_energy_instantaneous__hartree>{})\s+(?P<x_cp2k_md_potential_energy_average__hartree>{})".format(self.regexs.float, self.regexs.float)),
                        SM( " KINETIC ENERGY \[hartree\]     =\s+(?P<x_cp2k_md_kinetic_energy_instantaneous__hartree>{})\s+(?P<x_cp2k_md_kinetic_energy_average__hartree>{})".format(self.regexs.float, self.regexs.float)),
                        SM( " TEMPERATURE \[K\]              =\s+(?P<x_cp2k_md_temperature_instantaneous>{})\s+(?P<x_cp2k_md_temperature_average>{})".format(self.regexs.float, self.regexs.float)),
                        SM( " PRESSURE \[bar\]               =\s+(?P<x_cp2k_md_pressure_instantaneous__bar>{0})\s+(?P<x_cp2k_md_pressure_average__bar>{0})".format(self.regexs.float)),
                        SM( " BAROSTAT TEMP\[K\]             =\s+(?P<x_cp2k_md_barostat_temperature_instantaneous>{0})\s+(?P<x_cp2k_md_barostat_temperature_average>{0})".format(self.regexs.float)),
                        SM( " VOLUME\[bohr\^3\]               =\s+(?P<x_cp2k_md_volume_instantaneous__bohr3>{0})\s+(?P<x_cp2k_md_volume_average__bohr3>{0})".format(self.regexs.float)),
                        SM( " CELL LNTHS\[bohr\]             =\s+(?P<x_cp2k_md_cell_length_a_instantaneous__bohr>{0})\s+(?P<x_cp2k_md_cell_length_b_instantaneous__bohr>{0})\s+(?P<x_cp2k_md_cell_length_c_instantaneous__bohr>{0})".format(self.regexs.float)),
                        SM( " AVE. CELL LNTHS\[bohr\]        =\s+(?P<x_cp2k_md_cell_length_a_average__bohr>{0})\s+(?P<x_cp2k_md_cell_length_b_average__bohr>{0})\s+(?P<x_cp2k_md_cell_length_c_average__bohr>{0})".format(self.regexs.float)),
                        SM( " CELL ANGLS\[deg\]              =\s+(?P<x_cp2k_md_cell_angle_a_instantaneous__deg>{0})\s+(?P<x_cp2k_md_cell_angle_b_instantaneous__deg>{0})\s+(?P<x_cp2k_md_cell_angle_c_instantaneous__deg>{0})".format(self.regexs.float)),
                        SM( " AVE. CELL ANGLS\[deg\]         =\s+(?P<x_cp2k_md_cell_angle_a_average__deg>{0})\s+(?P<x_cp2k_md_cell_angle_b_average__deg>{0})\s+(?P<x_cp2k_md_cell_angle_c_average__deg>{0})".format(self.regexs.float)),
                    ],
                    adHoc=self.adHoc_save_md_quickstep()
                ),
            ]
        )

        # Compose root matcher according to the run type. This way the
        # unnecessary regex parsers will not be compiled and searched. Saves
        # computational time.
        self.root_matcher = SM("",
            forwardMatch=True,
            sections=["section_run"],
            subMatchers=[
                SM( "",
                    forwardMatch=True,
                    sections=["section_method"],
                    subMatchers=[
                        self.cm.restart(),
                        self.cm.header(),
                        self.cm.quickstep_header(),
                    ],
                ),
                self.md,
                self.cm.footer(),
            ]
        )

    #===========================================================================
    # onClose triggers
    def onClose_x_cp2k_section_md_settings(self, backend, gIndex, section):

        # Ensemble
        cp2k_ensemble = section.get_latest_value("x_cp2k_md_ensemble_type")
        self.cp2k_ensemble = cp2k_ensemble
        if cp2k_ensemble is not None:
            sampling_map = {
                "NVE": "NVE",
                "NVT": "NVT",
                "NPT_F": "NPT",
                "NPT_I": "NPT",
            }
            sampling = sampling_map.get(cp2k_ensemble)
            if sampling is not None:
                self.ensemble = sampling
                backend.addValue("ensemble_type", sampling)
            else:
                logger.error("Unknown ensemble type {}.".format(cp2k_ensemble))

        # Sampling type
        backend.addValue("sampling_method", "molecular_dynamics")

        # Print frequencies
        self.output_freq = section.get_latest_value("x_cp2k_md_print_frequency")
        self.energy_freq = section.get_latest_value("x_cp2k_md_energies_print_frequency")
        self.coord_freq = section.get_latest_value("x_cp2k_md_coordinates_print_frequency")
        self.velo_freq = section.get_latest_value("x_cp2k_md_velocities_print_frequency")
        self.cell_freq = section.get_latest_value("x_cp2k_md_simulation_cell_print_frequency")

        # Number of steps
        self.n_steps = section.get_latest_value("x_cp2k_md_number_of_time_steps")

        # Files
        coord_filename = section.get_latest_value("x_cp2k_md_coordinates_filename")
        vel_filename = section.get_latest_value("x_cp2k_md_velocities_filename")
        energies_filename = section.get_latest_value("x_cp2k_md_energies_filename")
        cell_filename = section.get_latest_value("x_cp2k_md_simulation_cell_filename")
        coord_filepath = self.file_service.set_file_id(coord_filename, "coordinates")
        vel_filepath = self.file_service.set_file_id(vel_filename, "velocities")
        cell_filepath = self.file_service.set_file_id(cell_filename, "cell")
        energies_filepath = self.file_service.set_file_id(energies_filename, "energies")

        # Setup trajectory iterator
        traj_format = self.cache_service["trajectory_format"]
        if traj_format is not None and coord_filepath is not None:

            # Use special parsing for CP2K pdb files because they don't follow the proper syntax
            if traj_format == "PDB":
                self.traj_iterator = nomadcore.csvparsing.iread(coord_filepath, columns=[3, 4, 5], start="CRYST", end="END")
            else:
                try:
                    self.traj_iterator = nomadcore.configurationreading.iread(coord_filepath)
                except ValueError:
                    pass

        # Setup velocity iterator
        vel_format = self.cache_service["velocity_format"]
        if vel_format is not None and vel_filepath is not None:
            try:
                self.vel_iterator = nomadcore.configurationreading.iread(vel_filepath)
            except ValueError:
                pass

        # Setup energy file iterator
        if energies_filepath is not None:
            self.energy_iterator = nomadcore.csvparsing.iread(energies_filepath, columns=[0, 1, 2, 3, 4, 5, 6], comments="#")

        # Setup cell file iterator
        if cell_filepath is not None:
            self.cell_iterator = nomadcore.csvparsing.iread(cell_filepath, columns=[2, 3, 4, 5, 6, 7, 8, 9, 10], comments="#")

    def onClose_x_cp2k_section_md(self, backend, gIndex, section):

        # Determine the highest print frequency and use that as the number of
        # single configuration calculations
        freqs = {
            "output": [self.output_freq, True],
            "trajectory": [self.coord_freq, True],
            "velocities": [self.velo_freq, True],
            "energies": [self.energy_freq, True],
            "cell": [self.cell_freq, True],
        }

        # See if the files actually exist
        traj_file = self.file_service.get_file_by_id("coordinates")
        if traj_file is None:
            freqs["trajectory"][1] = False
        velocities_file = self.file_service.get_file_by_id("velocities")
        if velocities_file is None:
            freqs["velocities"][1] = False
        energies_file = self.file_service.get_file_by_id("energies")
        if energies_file is None:
            freqs["energies"][1] = False
        cell_file = self.file_service.get_file_by_id("cell")
        if cell_file is None:
            freqs["cell"][1] = False

        # See if we can determine the units
        traj_unit = self.cache_service["trajectory_unit"]
        if traj_unit is None:
            freqs["coordinates"][1] = False
        vel_unit = self.cache_service["velocity_unit"]
        if vel_unit is None:
            freqs["velocities"][1] = False

        # Trajectory print settings
        add_last_traj = False
        add_last_traj_setting = self.cache_service["traj_add_last"]
        if add_last_traj_setting == "NUMERIC" or add_last_traj_setting == "SYMBOLIC":
            add_last_traj = True

        # Velocities print settings
        add_last_vel = False
        add_last_vel_setting = self.cache_service["vel_add_last"]
        if add_last_vel_setting == "NUMERIC" or add_last_vel_setting == "SYMBOLIC":
            add_last_vel = True

        last_step = self.n_steps
        md_steps = section["x_cp2k_section_md_step"]

        frame_sequence_potential_energy = []
        frame_sequence_temperature = []
        frame_sequence_time = []
        frame_sequence_kinetic_energy = []
        frame_sequence_conserved_quantity = []
        frame_sequence_pressure = []
        ener_frames = []

        single_conf_gids = []
        i_md_step = 0

        # Check that is the calculation starting from first frame. If not, this
        # is a restarted calculation. In this case it is practically impossible
        # to know from which frame number we should start reading the
        # configurations, because the print settings in the previous runs may
        # have been different from now, and the step numbers are hard to align.
        # In this case we just parse what is found in this output file. The
        # REFTRAJ calculations use a different indexing that starts from 2
        # instead of one.
        start_step_number = md_steps[1]["x_cp2k_md_step_number"][0]
        if (self.cp2k_ensemble.upper() == "REFTRAJ" and start_step_number != 2) or \
           (self.cp2k_ensemble.upper() != "REFTRAJ" and start_step_number != 1):
            self.traj_iterator = None
            self.cell_iterator = None
            self.vel_iterator = None
            self.energy_iterator = None

        # The reftraj mode has axactly the requested number of steps, but other
        # modes have one more.
        if self.cp2k_ensemble.upper() == "REFTRAJ":
            n_steps = self.n_steps
        else:
            n_steps = self.n_steps + 1

        for i_step in range(n_steps):

            sectionGID = backend.openSection("section_single_configuration_calculation")
            systemGID = backend.openSection("section_system")
            single_conf_gids.append(sectionGID)

            # If NPT is run, and the cell file is not available, output the
            # simulation cel only on the first step to section_system
            if i_step == 0 and self.ensemble == "NPT" and self.cell_iterator is None:
                self.cache_service.addArrayValues("simulation_cell", unit="angstrom")
                self.cache_service.addArrayValues("lattice_vectors", unit="angstrom")

            # Unchanging cell
            if self.ensemble != "NPT":
                self.cache_service.addArrayValues("simulation_cell", unit="angstrom")
                self.cache_service.addArrayValues("lattice_vectors", unit="angstrom")

            # Trajectory
            if freqs["trajectory"][1] and self.traj_iterator is not None:
                if (i_step) % freqs["trajectory"][0] == 0 or (i_step == last_step and add_last_traj):
                    try:
                        pos = next(self.traj_iterator)
                    except StopIteration:
                        logger.error("Could not get the next geometries from an external file. It seems that the number of optimization steps in the CP2K outputfile doesn't match the number of steps found in the external trajectory file.")
                    else:
                        backend.addArrayValues("atom_positions", pos, unit=traj_unit)

            # Velocities
            if freqs["velocities"][1] and self.vel_iterator is not None:
                if (i_step) % freqs["velocities"][0] == 0 or (i_step == last_step and add_last_vel):
                    try:
                        vel = next(self.vel_iterator)
                    except StopIteration:
                        logger.error("Could not get the next velociies from an external file. It seems that the number of optimization steps in the CP2K outpufile doesn't match the number of steps found in the external velocities file.")
                    else:
                        backend.addArrayValues("atom_velocities", vel, unit=vel_unit)

            # Energy file
            if self.energy_iterator is not None:
                if (i_step) % freqs["energies"][0] == 0:
                    line = next(self.energy_iterator)

                    time = line[1]
                    kinetic_energy = line[2]
                    temperature = line[3]
                    potential_energy = line[4]
                    conserved_quantity = line[5]
                    wall_time = line[6]

                    frame_sequence_time.append(time)
                    frame_sequence_kinetic_energy.append(kinetic_energy)
                    frame_sequence_temperature.append(temperature)
                    frame_sequence_potential_energy.append(potential_energy)
                    frame_sequence_conserved_quantity.append(conserved_quantity)

                    ener_frames.append(i_step)

                    backend.addValue("energy_total", conserved_quantity)
                    backend.addValue("time_calculation", wall_time)

            # Cell file
            if self.cell_iterator is not None:
                if (i_step) % freqs["cell"][0] == 0:
                    line = next(self.cell_iterator)
                    cell = np.reshape(line, (3, 3))
                    self.backend.addArrayValues("simulation_cell", cell, unit="angstrom")
                    self.backend.addArrayValues("lattice_vectors", cell, unit="angstrom")

            # Output file
            if md_steps:
                if (i_step) % freqs["output"][0] == 0:
                    try:
                        md_step = md_steps[i_md_step]

                    except IndexError:
                        # The process has stopped unexpectedly, but still we
                        # should report all the steps so far. So we break the
                        # loop and retain what was parsed.
                        break
                    # print(md_step["x_cp2k_md_step_number"])
                    quickstep = self.md_quicksteps[i_md_step]
                    if quickstep is not None:
                        quickstep.add_latest_value("x_cp2k_atom_forces", "atom_forces")
                        quickstep.add_latest_value("x_cp2k_stress_tensor", "stress_tensor")
                        scfGID = backend.openSection("section_scf_iteration")
                        quickstep.add_latest_value(["x_cp2k_section_scf_iteration", "x_cp2k_energy_total_scf_iteration"], "energy_total_scf_iteration")
                        quickstep.add_latest_value(["x_cp2k_section_scf_iteration", "x_cp2k_energy_change_scf_iteration"], "energy_change_scf_iteration")
                        quickstep.add_latest_value(["x_cp2k_section_scf_iteration", "x_cp2k_energy_XC_scf_iteration"], "energy_XC_scf_iteration")
                        backend.closeSection("section_scf_iteration", scfGID)
                    i_md_step += 1
                    pressure = md_step.get_latest_value("x_cp2k_md_pressure_instantaneous")
                    if pressure is not None:
                        frame_sequence_pressure.append(pressure)

            backend.closeSection("section_system", systemGID)
            backend.closeSection("section_single_configuration_calculation", sectionGID)

        # Add the summaries to frame sequence
        if len(frame_sequence_potential_energy) != 0:
            frame_sequence_potential_energy = convert_unit(np.array(frame_sequence_potential_energy), "hartree")
            backend.addArrayValues("frame_sequence_potential_energy", frame_sequence_potential_energy)
            backend.addArrayValues("frame_sequence_potential_energy_frames", np.array(ener_frames))
            mean_pot = frame_sequence_potential_energy.mean()
            std_pot = frame_sequence_potential_energy.std()
            backend.addArrayValues("frame_sequence_potential_energy_stats", np.array([mean_pot, std_pot]))
        if len(frame_sequence_kinetic_energy) != 0:
            frame_sequence_kinetic_energy = convert_unit(np.array(frame_sequence_kinetic_energy), "hartree")
            backend.addArrayValues("frame_sequence_kinetic_energy", frame_sequence_kinetic_energy)
            backend.addArrayValues("frame_sequence_kinetic_energy_frames", np.array(ener_frames))
            mean_kin = frame_sequence_kinetic_energy.mean()
            std_kin = frame_sequence_kinetic_energy.std()
            backend.addArrayValues("frame_sequence_kinetic_energy_stats", np.array([mean_kin, std_kin]))
        if len(frame_sequence_conserved_quantity) != 0:
            frame_sequence_conserved_quantity = convert_unit(np.array(frame_sequence_conserved_quantity), "hartree")
            backend.addArrayValues("frame_sequence_conserved_quantity", frame_sequence_conserved_quantity)
            backend.addArrayValues("frame_sequence_conserved_quantity_frames", np.array(ener_frames))
            mean_cons = frame_sequence_conserved_quantity.mean()
            std_cons = frame_sequence_conserved_quantity.std()
            backend.addArrayValues("frame_sequence_conserved_quantity_stats", np.array([mean_cons, std_cons]))
        if len(frame_sequence_time) != 0:
            frame_sequence_time = np.array(frame_sequence_time)
            backend.addArrayValues("frame_sequence_time", frame_sequence_time, unit="fs")
        if len(frame_sequence_temperature) != 0:
            frame_sequence_temperature = np.array(frame_sequence_temperature)
            backend.addArrayValues("frame_sequence_temperature", frame_sequence_temperature)
            backend.addArrayValues("frame_sequence_temperature_frames", np.array(ener_frames))
            mean_temp = frame_sequence_temperature.mean()
            std_temp = frame_sequence_temperature.std()
            backend.addArrayValues("frame_sequence_temperature_stats", np.array([mean_temp, std_temp]))
        if len(frame_sequence_pressure) != 0:
            frame_sequence_pressure = np.array(frame_sequence_pressure)
            backend.addArrayValues("frame_sequence_pressure", frame_sequence_pressure)
            mean_pressure = frame_sequence_pressure.mean()
            std_pressure = frame_sequence_pressure.std()
            backend.addArrayValues("frame_sequence_pressure_stats", np.array([mean_pressure, std_pressure]))

        # Number of frames. We open a SCC for each frame, even if there is no
        # information for it.
        backend.addValue("number_of_frames_in_sequence", len(single_conf_gids))

        # Reference to sampling method
        backend.addValue("frame_sequence_to_sampling_ref", 0)

        # References to local frames
        backend.addArrayValues("frame_sequence_local_frames_ref", np.array(single_conf_gids))

    #===========================================================================
    # adHoc functions
    def adHoc_save_md_quickstep(self):
        def wrapper(parser):
            section_managers = parser.backend.sectionManagers
            section_run_manager = section_managers["section_run"]
            section_run = section_run_manager.openSections[0]
            quickstep = section_run.get_latest_value("x_cp2k_section_quickstep_calculation")
            self.md_quicksteps.append(quickstep)
        return wrapper



from typing import Union

from fastapi import APIRouter, Request

from optimade import __api_version__

from optimade.models import (
    ErrorResponse,
    IndexInfoResponse,
    IndexInfoAttributes,
    IndexInfoResource,
    IndexRelationship,
)

from optimade.server.config import CONFIG

from .utils import meta_values


router = APIRouter(redirect_slashes=True)


@router.get(
    "/info",
    response_model=Union[IndexInfoResponse, ErrorResponse],
    response_model_exclude_unset=True,
    tags=["Info"],
)
def get_info(request: Request):
    return IndexInfoResponse(
        meta=meta_values(str(request.url), 1, 1, more_data_available=False),
        data=IndexInfoResource(
            id=IndexInfoResource.schema()["properties"]["id"]["const"],
            type=IndexInfoResource.schema()["properties"]["type"]["const"],
            attributes=IndexInfoAttributes(
                api_version=f"v{__api_version__}",
                available_api_versions=[
                    {
                        "url": f"{CONFIG.provider.index_base_url}/v{__api_version__.split('.')[0]}/",
                        "version": f"{__api_version__}",
                    }
                ],
                formats=["json"],
                available_endpoints=["info", "links"],
                entry_types_by_format={"json": []},
                is_index=True,
            ),
            relationships={
                "default": IndexRelationship(
                    data={"type": "child", "id": CONFIG.default_db}
                )
            },
        ),
    )



import numpy as np

class Grid():
    
    @staticmethod
    def create(grid_id = None, grid_type = 'dg_cut', mu = -2, sigma = 7, cutoff = (-10,5), num_bins = 56, original_stepsize = 0.05):
        if grid_id == None:
            self = Grid()
            self.grid_type = grid_type
            self.mu = mu
            self.sigma = sigma
            self.cutoff = cutoff
            self.num_bins = num_bins
            self.original_stepsize = original_stepsize
            self.grid_id = self.get_grid_id()
        else:
            self = Grid().create(**Grid().resolve_grid_id(grid_id))
        return self

    def get_grid_id(self):
        id = ':'.join([self.grid_type, str(self.num_bins), str(self.mu), str(self.sigma), str(self.cutoff)])
        return id

    @staticmethod
    def resolve_grid_id(grid_id):
        grid_type, num_bins, mu, sigma, cutoff = grid_id.split(':')
        return {'grid_type' : grid_type, 'num_bins' : int(num_bins), 'mu' : float(mu), 'sigma' : float(sigma), 'cutoff' : tuple([float(x) for x in cutoff[1:-1].split(',')])}

    def grid(self):
        asc = 0
        desc = 0
        x_grid = [0]
        while (asc is not None) or (desc is not None):
            if asc is not None:
                asc += self._step_sequencer(asc, self.mu, self.sigma, self.original_stepsize) * self.original_stepsize
                x_grid = x_grid + [round(asc, 8)]
                if asc > self.cutoff[1]:
                    asc = None
            if desc is not None:
                desc -= self._step_sequencer(desc, self.mu, self.sigma, self.original_stepsize) * self.original_stepsize
                x_grid = [round(desc, 8)] + x_grid
                if desc < self.cutoff[0]:
                    desc = None
        grid = []
        for item in x_grid:
            bins = []
            bin_height = self._step_sequencer(item, self.mu, self.sigma, original_stepsize=0.1) / (self.num_bins * 2.)
            for idx in range(1, self.num_bins + 1):
                bins.append(bin_height * idx)
            grid.append([item, bins])
        return grid

    def _gauss(self, x, mu, sigma, normalized=True):
        coefficient = (np.sqrt(2 * np.pi) * sigma)
        value = np.exp((-0.5) * ((x - mu) / (sigma * 1.)) ** 2)
        if normalized:
            value = value / (coefficient * 1.)
        return value

    def _step_sequencer(self, x, mu, sigma, original_stepsize):
        return int(round((1 + original_stepsize - self._gauss(x, mu, sigma, normalized=False)) / original_stepsize, 9))



import os
import numpy as np
import phonopy
from phonopy.units import THzToEv

from phonopyparser.fhiaims_io import Control, read_forces_aims
from phonopyparser.phonopy_properties import PhononProperties
from phonopyparser.FHIaims import read_aims

from nomadcore.unit_conversion.unit_conversion import convert_unit_function

from nomad.datamodel.metainfo.public import section_run, section_system,\
    section_system_to_system_refs, section_method, section_single_configuration_calculation,\
    section_k_band, section_k_band_segment, section_dos, section_frame_sequence,\
    section_thermodynamical_properties, section_sampling_method, Workflow, Phonon, \
    section_calculation_to_calculation_refs


class PhonopyCalculatorInterface:
    def __init__(self, filepath, archive, logger=None):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logger
        self._phonopy_obj = None
        self.calculator = None
        if 'control.in' in filepath:
            self.calculator = 'fhi-aims'
        elif 'phonon.yaml' in filepath:
            self.calculator = 'vasp'
        self.references = []

    @property
    def phonopy_obj(self):
        if self._phonopy_obj is None:
            if self.calculator == 'fhi-aims':
                self._build_phonopy_object_fhi_aims()
            elif self.calculator == 'vasp':
                self._build_phonopy_object_vasp()
        return self._phonopy_obj

    def _build_phonopy_object_vasp(self):
        cwd = os.getcwd()
        os.chdir(os.path.dirname(self.filepath))

        phonopy_obj = phonopy.load('phonon.yaml')
        os.chdir(cwd)

        self._phonopy_obj = phonopy_obj

    def _build_phonopy_object_fhi_aims(self):
        cwd = os.getcwd()
        os.chdir(os.path.dirname(os.path.dirname(self.filepath)))

        cell_obj = read_aims("geometry.in")
        control = Control()
        if (len(control.phonon["supercell"]) == 3):
            supercell_matrix = np.diag(control.phonon["supercell"])
        elif (len(control.phonon["supercell"]) == 9):
            supercell_matrix = np.array(control.phonon["supercell"]).reshape(3, 3)
        displacement = control.phonon["displacement"]
        sym = control.phonon["symmetry_thresh"]

        set_of_forces, phonopy_obj, Relative_Path = read_forces_aims(
            cell_obj, supercell_matrix, displacement, sym)
        # Prep_Path = name.split("phonopy-FHI-aims-displacement-")

        phonopy_obj.set_forces(set_of_forces)
        phonopy_obj.produce_force_constants()

        os.chdir(cwd)

        self._phonopy_obj = phonopy_obj

    def parse_bandstructure(self):
        freqs, bands, bands_labels = self.properties.get_bandstructure()

        # convert THz to eV
        freqs = freqs * THzToEv

        # convert eV to J
        eVtoJoules = convert_unit_function('eV', 'joules')
        freqs = eVtoJoules(freqs)

        sec_scc = self.archive.section_run[0].section_single_configuration_calculation[0]

        sec_k_band = sec_scc.m_create(section_k_band)
        sec_k_band.band_structure_kind = 'vibrational'

        for i in range(len(freqs)):
            freq = np.expand_dims(freqs[i], axis=0)
            sec_k_band_segment = sec_k_band.m_create(section_k_band_segment)
            sec_k_band_segment.band_energies = freq
            sec_k_band_segment.band_k_points = bands[i]
            sec_k_band_segment.band_segm_labels = bands_labels[i]

    def parse_dos(self):
        f, dos = self.properties.get_dos()

        # To match the shape given in meta data another dimension is added to the
        # array (spin degress of fredom is 1)
        dos = np.expand_dims(dos, axis=0)

        # convert THz to eV to Joules
        eVtoJoules = convert_unit_function('eV', 'joules')
        f = f * THzToEv
        f = eVtoJoules(f)

        sec_scc = self.archive.section_run[0].section_single_configuration_calculation[0]
        sec_dos = sec_scc.m_create(section_dos)
        sec_dos.dos_kind = 'vibrational'
        sec_dos.dos_values = dos
        sec_dos.dos_energies = f

    def parser_thermodynamical_properties(self):
        T, fe, _, cv = self.properties.get_thermodynamical_properties()

        n_atoms = len(self.phonopy_obj.unitcell)
        n_atoms_supercell = len(self.phonopy_obj.supercell)

        fe = fe / n_atoms

        # The thermodynamic properties are reported by phonopy for the base
        # system. Since the values in the metainfo are stored per the referenced
        # system, we need to multiple by the size factor between the base system
        # and the supersystem used in the calculations.
        cv = cv * (n_atoms_supercell / n_atoms)

        # convert to SI units
        eVtoJoules = convert_unit_function('eV', 'joules')
        eVperKtoJoules = convert_unit_function('eV*K**-1', 'joules*K**-1')
        fe = eVtoJoules(fe)
        cv = eVperKtoJoules(cv)

        sec_run = self.archive.section_run[0]
        sec_scc = sec_run.section_single_configuration_calculation

        sec_frame_sequence = sec_run.m_create(section_frame_sequence)
        sec_frame_sequence.frame_sequence_local_frames_ref = sec_scc

        sec_thermo_prop = sec_frame_sequence.m_create(section_thermodynamical_properties)
        sec_thermo_prop.thermodynamical_property_temperature = T
        sec_thermo_prop.vibrational_free_energy_at_constant_volume = fe
        sec_thermo_prop.thermodynamical_property_heat_capacity_C_v = cv

        sec_sampling_method = sec_run.m_create(section_sampling_method)
        sec_sampling_method.sampling_method = 'taylor_expansion'
        sec_sampling_method.sampling_method_expansion_order = 2

        sec_frame_sequence.frame_sequence_to_sampling_ref = sec_sampling_method

    def parse_ref(self):
        sec_scc = self.archive.section_run[0].section_single_configuration_calculation[0]
        sec_calc_refs = sec_scc.m_create(section_calculation_to_calculation_refs)
        sec_calc_refs.calculation_to_calculation_kind = 'source_calculation'
        for ref in self.references:
            sec_calc_refs.calculation_to_calculation_external_url = ref

    def parse(self):
        phonopy_obj = self.phonopy_obj
        self.properties = PhononProperties(self.phonopy_obj, self.logger)

        pbc = np.array((1, 1, 1), bool)

        unit_cell = self.phonopy_obj.unitcell.get_cell()
        unit_pos = self.phonopy_obj.unitcell.get_positions()
        unit_sym = np.array(self.phonopy_obj.unitcell.get_chemical_symbols())

        super_cell = self.phonopy_obj.supercell.get_cell()
        super_pos = self.phonopy_obj.supercell.get_positions()
        super_sym = np.array(self.phonopy_obj.supercell.get_chemical_symbols())

        # convert to SI
        convert_fc = convert_unit_function('eV*angstrom**-2', 'joules*meter**-2')
        convert_angstrom = convert_unit_function('angstrom', 'meter')

        unit_cell = convert_angstrom(unit_cell)
        unit_pos = convert_angstrom(unit_pos)

        super_cell = convert_angstrom(super_cell)
        super_pos = convert_angstrom(super_pos)

        displacement = np.linalg.norm(phonopy_obj.displacements[0][1:])
        displacement = convert_angstrom(displacement)
        supercell_matrix = phonopy_obj.supercell_matrix
        sym_tol = phonopy_obj.symmetry.tolerance
        force_constants = phonopy_obj.get_force_constants()
        force_constants = convert_fc(force_constants)

        sec_run = self.archive.m_create(section_run)
        sec_run.program_name = 'Phonopy'
        sec_run.program_version = phonopy.__version__
        sec_system_unit = sec_run.m_create(section_system)
        sec_system_unit.configuration_periodic_dimensions = pbc
        sec_system_unit.atom_labels = unit_sym
        sec_system_unit.atom_positions = unit_pos
        sec_system_unit.simulation_cell = unit_cell

        sec_system = sec_run.m_create(section_system)
        sec_system_to_system_refs = sec_system.m_create(section_system_to_system_refs)
        sec_system_to_system_refs.system_to_system_kind = 'subsystem'
        sec_system_to_system_refs.system_to_system_ref = sec_system_unit
        sec_system.configuration_periodic_dimensions = pbc
        sec_system.atom_labels = super_sym
        sec_system.atom_positions = super_pos
        sec_system.simulation_cell = super_cell
        sec_system.SC_matrix = supercell_matrix
        sec_system.x_phonopy_original_system_ref = sec_system_unit

        sec_method = sec_run.m_create(section_method)
        sec_method.x_phonopy_symprec = sym_tol
        sec_method.x_phonopy_displacement = displacement

        sec_scc = sec_run.m_create(section_single_configuration_calculation)
        sec_scc.single_configuration_calculation_to_system_ref = sec_system
        sec_scc.single_configuration_to_calculation_method_ref = sec_method
        sec_scc.hessian_matrix = force_constants

        self.parse_bandstructure()
        self.parse_dos()
        self.parser_thermodynamical_properties()
        self.parse_ref()

        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.workflow_type = 'phonon'
        sec_phonon = sec_workflow.m_create(Phonon)
        sec_phonon.force_calculator = self.calculator
        vol = np.dot(unit_cell[0], np.cross(unit_cell[1], unit_cell[2]))
        sec_phonon.mesh_density = np.prod(self.properties.mesh) / vol
        n_imaginary = np.count_nonzero(self.properties.frequencies < 0)
        sec_phonon.n_imaginary_frequencies = n_imaginary
        if phonopy_obj.nac_params:
            sec_phonon.with_non_analytic_correction = True



# pylint: disable=invalid-name
from typing import List, Tuple, Iterable

from optimade.models.structures import Vector3D

try:
    import numpy as np
except ImportError:
    from warnings import warn

    np = None
    NUMPY_NOT_FOUND = "NumPy not found, cannot convert structure to your desired format"


def scaled_cell(
    cell: Tuple[Vector3D, Vector3D, Vector3D]
) -> Tuple[Vector3D, Vector3D, Vector3D]:
    """Return a scaled 3x3 cell from cartesian 3x3 cell (`lattice_vectors`)

    This is based on PDB's method of calculating SCALE from CRYST data.
    For more info, see https://www.wwpdb.org/documentation/file-format-content/format33/sect8.html#SCALEn
    """
    if globals().get("np", None) is None:
        warn(NUMPY_NOT_FOUND)
        return None

    cell = np.asarray(cell)

    volume = np.dot(cell[0], np.cross(cell[1], cell[2]))
    scale = []
    for i in range(3):
        vector = np.cross(cell[(i + 1) % 3], cell[(i + 2) % 3]) / volume
        scale.append(tuple(vector))
    return tuple(scale)


def fractional_coordinates(
    cell: Tuple[Vector3D, Vector3D, Vector3D], cartesian_positions: List[Vector3D]
) -> List[Vector3D]:
    """Returns fractional coordinates and wraps coordinates to be [0;1[

    NOTE: Based on `ase.atoms:Atoms.get_scaled_positions()`.
    """
    if globals().get("np", None) is None:
        warn(NUMPY_NOT_FOUND)
        return None

    cell = np.asarray(cell)
    cartesian_positions = np.asarray(cartesian_positions)

    fractional = np.linalg.solve(cell.T, cartesian_positions.T).T

    # Expecting a bulk 3D structure here, note, this may change in the future.
    # See `ase.atoms:Atoms.get_scaled_positions()` for ideas on how to handle lower dimensional structures.
    # Furthermore, according to ASE we need to modulo 1.0 twice.
    # This seems to be due to small floats % 1.0 becomes 1.0, hence twice makes it 0.0.
    for i in range(3):
        fractional[:, i] %= 1.0
        fractional[:, i] %= 1.0

    return [tuple(position) for position in fractional]


def cell_to_cellpar(cell, radians=False):
    """Returns the cell parameters [a, b, c, alpha, beta, gamma].

    Angles are in degrees unless radian=True is used.

    NOTE: Based on `ase.geometry.cell:cell_to_cellpar()`.
    """
    if globals().get("np", None) is None:
        warn(NUMPY_NOT_FOUND)
        return None

    cell = np.asarray(cell)

    lengths = [np.linalg.norm(vector) for vector in cell]
    angles = []
    for i in range(3):
        j = i - 1
        k = i - 2
        outer_product = lengths[j] * lengths[k]
        if outer_product > 1e-16:
            x_vector = np.dot(cell[j], cell[k]) / outer_product
            angle = 180.0 / np.pi * np.arccos(x_vector)
        else:
            angle = 90.0
        angles.append(angle)
    if radians:
        angles = [angle * np.pi / 180 for angle in angles]
    return np.array(lengths + angles)


def unit_vector(x):
    """Return a unit vector in the same direction as x."""
    if globals().get("np", None) is None:
        warn(NUMPY_NOT_FOUND)
        return None

    y = np.array(x, dtype="float")
    return y / np.linalg.norm(y)


def cellpar_to_cell(cellpar, ab_normal=(0, 0, 1), a_direction=None):
    """Return a 3x3 cell matrix from cellpar=[a,b,c,alpha,beta,gamma].

    Angles must be in degrees.

    The returned cell is orientated such that a and b
    are normal to `ab_normal` and a is parallel to the projection of
    `a_direction` in the a-b plane.

    Default `a_direction` is (1,0,0), unless this is parallel to
    `ab_normal`, in which case default `a_direction` is (0,0,1).

    The returned cell has the vectors va, vb and vc along the rows. The
    cell will be oriented such that va and vb are normal to `ab_normal`
    and va will be along the projection of `a_direction` onto the a-b
    plane.

    Example:

    >>> cell = cellpar_to_cell([1, 2, 4, 10, 20, 30], (0, 1, 1), (1, 2, 3))
    >>> np.round(cell, 3)
    array([[ 0.816, -0.408,  0.408],
           [ 1.992, -0.13 ,  0.13 ],
           [ 3.859, -0.745,  0.745]])

    NOTE: Direct copy of `ase.geometry.cell:cellpar_to_cell()`.
    """
    if globals().get("np", None) is None:
        warn(NUMPY_NOT_FOUND)
        return None

    if a_direction is None:
        if np.linalg.norm(np.cross(ab_normal, (1, 0, 0))) < 1e-5:
            a_direction = (0, 0, 1)
        else:
            a_direction = (1, 0, 0)

    # Define rotated X,Y,Z-system, with Z along ab_normal and X along
    # the projection of a_direction onto the normal plane of Z.
    a_direction_array = np.array(a_direction)
    Z = unit_vector(ab_normal)
    X = unit_vector(a_direction_array - np.dot(a_direction_array, Z) * Z)
    Y = np.cross(Z, X)

    # Express va, vb and vc in the X,Y,Z-system
    alpha, beta, gamma = 90.0, 90.0, 90.0
    if isinstance(cellpar, (int, float)):
        a = b = c = cellpar
    elif len(cellpar) == 1:
        a = b = c = cellpar[0]
    elif len(cellpar) == 3:
        a, b, c = cellpar
    else:
        a, b, c, alpha, beta, gamma = cellpar

    # Handle orthorhombic cells separately to avoid rounding errors
    eps = 2 * np.spacing(90.0, dtype=np.float64)  # around 1.4e-14
    # alpha
    if abs(abs(alpha) - 90) < eps:
        cos_alpha = 0.0
    else:
        cos_alpha = np.cos(alpha * np.pi / 180.0)
    # beta
    if abs(abs(beta) - 90) < eps:
        cos_beta = 0.0
    else:
        cos_beta = np.cos(beta * np.pi / 180.0)
    # gamma
    if abs(gamma - 90) < eps:
        cos_gamma = 0.0
        sin_gamma = 1.0
    elif abs(gamma + 90) < eps:
        cos_gamma = 0.0
        sin_gamma = -1.0
    else:
        cos_gamma = np.cos(gamma * np.pi / 180.0)
        sin_gamma = np.sin(gamma * np.pi / 180.0)

    # Build the cell vectors
    va = a * np.array([1, 0, 0])
    vb = b * np.array([cos_gamma, sin_gamma, 0])
    cx = cos_beta
    cy = (cos_alpha - cos_beta * cos_gamma) / sin_gamma
    cz_sqr = 1.0 - cx * cx - cy * cy
    assert cz_sqr >= 0
    cz = np.sqrt(cz_sqr)
    vc = c * np.array([cx, cy, cz])

    # Convert to the Cartesian x,y,z-system
    abc = np.vstack((va, vb, vc))
    T = np.vstack((X, Y, Z))
    cell = np.dot(abc, T)

    return cell


def _pad_iter_of_iters(
    iterable: Iterable[Iterable],
    padding: float = None,
    outer: Iterable = None,
    inner: Iterable = None,
) -> Tuple[Iterable[Iterable], bool]:
    """Turn any null/None values into a float in given iterable of iterables"""
    try:
        padding = float(padding)
    except (TypeError, ValueError):
        padding = float("nan")

    outer = outer if outer is not None else list
    inner = inner if outer is not None else tuple

    padded_iterable = any(
        value is None for inner_iterable in iterable for value in inner_iterable
    )

    if padded_iterable:
        padded_iterable_of_iterables = []
        for inner_iterable in iterable:
            new_inner_iterable = inner(
                value if value is not None else padding for value in inner_iterable
            )
            padded_iterable_of_iterables.append(new_inner_iterable)
        iterable = outer(padded_iterable_of_iterables)

    return iterable, padded_iterable


def pad_positions(
    positions: List[Vector3D], padding: float = None
) -> Tuple[List[Vector3D], bool]:
    """Turn any null/None values into a float in given list of positions"""
    return _pad_iter_of_iters(
        iterable=positions, padding=padding, outer=list, inner=tuple,
    )


def pad_cell(
    lattice_vectors: Tuple[Vector3D, Vector3D, Vector3D], padding: float = None
) -> Tuple[Tuple[Vector3D, Vector3D, Vector3D], bool]:
    """Turn any null/None values into a float in given tuple of lattice_vectors"""
    return _pad_iter_of_iters(
        iterable=lattice_vectors, padding=padding, outer=tuple, inner=tuple,
    )



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# import setup_paths
# import os, sys, json
import logging

from nomadcore.baseclasses import ParserInterface, ParserContext
from .LammpsLogParser import LammpsMainParser

# from nomadcore.simple_parser import SimpleMatcher, mainFunction
# from LammpsParserCommon import get_metaInfo, get_parseInfo

logger = logging.getLogger(name="nomad.LammpsParser")


class LammpsContext(ParserContext):
    def __init__(self):
        super(LammpsContext, self).__init__()


class LammpsParser(ParserInterface):
    """This class handles the initial setup before any parsing can happen. It
    determines which version of the software was used to generate the output
    and then sets up a correct main parser.

    After the implementation has been setup, you can parse the files with
    parse().
    """
    def __init__(self, metainfo_to_keep=None, backend=None, default_units=None, metainfo_units=None):
        super(LammpsParser, self).__init__(metainfo_to_keep, backend, default_units, metainfo_units)

    def setup_version(self):
        """Setups the version by looking at the output file and the version
        specified in it.
        """
        # Search for the version specification and initialize a correct
        # main parser for this version.

        self.main_parser = LammpsMainParser(self.parser_context.main_file, self.parser_context)

    def get_metainfo_filename(self):
        return "lammps.nomadmetainfo.json"

    def get_parser_info(self):
        return {'name': 'lammps-parser', 'version': '1.0'}



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public
from nomad.datamodel.metainfo import common

m_package = Package(
    name='castep_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='castep.nomadmetainfo.json'))


class x_castep_section_vibrational_frequencies(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_vibrational_frequencies'))

    x_castep_vibrationl_frequencies = Quantity(
        type=np.dtype(np.float64),
        shape=['len(self.nr_iter)'],
        description='''
        Vibration Frequenices (cm-1)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_vibrationl_frequencies'))

    x_castep_vibrationl_frequencies_store = Quantity(
        type=str,
        shape=[],
        description='''
        Vibration Frequenices (cm-1)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_vibrationl_frequencies_store'))

    x_castep_ir_store = Quantity(
        type=str,
        shape=[],
        description='''
        Irreducible representation in the Point Group
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ir_store'))

    x_castep_ir = Quantity(
        type=str,
        shape=['len(self.nr_iter)'],
        description='''
        Irreducible representation in the Point Group
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ir'))

    x_castep_raman_activity = Quantity(
        type=np.dtype(np.float64),
        shape=['len(self.nr_iter)'],
        description='''
        Raman activity (A**4/amu)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_raman_activity'))

    x_castep_raman_active = Quantity(
        type=str,
        shape=[],
        description='''
        Raman activity (A**4/amu)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_raman_active'))

    x_castep_n_iterations_phonons = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of iterations in phonons
        ''',
        a_legacy=LegacyDefinition(name='x_castep_n_iterations_phonons'))

    x_castep_raman_activity_store = Quantity(
        type=str,
        shape=[],
        description='''
        Raman activity (A**4/amu)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_raman_activity_store'))

    x_castep_ir_intensity = Quantity(
        type=np.dtype(np.float64),
        shape=['len(self.nr_iter)'],
        description='''
        IR intensities (D/A)**2/amu
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ir_intensity'))

    x_castep_ir_intensity_store = Quantity(
        type=str,
        shape=[],
        description='''
        IR intensities (D/A)**2/amu
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ir_intensity_store'))


class x_castep_section_band_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_band_parameters'))

    x_castep_band_n_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_castep_band_n_bands
        ''',
        a_legacy=LegacyDefinition(name='x_castep_band_n_bands'))

    x_castep_band_conv_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_castep_band_conv_tolerance
        ''',
        a_legacy=LegacyDefinition(name='x_castep_band_conv_tolerance'))

    x_castep_band_n_iterations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_castep_band_n_iterations
        ''',
        a_legacy=LegacyDefinition(name='x_castep_band_n_iterations'))

    x_castep_band_max_cg = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_castep_band_max_cg
        ''',
        a_legacy=LegacyDefinition(name='x_castep_band_max_cg'))


class x_castep_section_core_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_core_parameters'))

    x_castep_core_spectra_n_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_castep_core_spectra_n_bands
        ''',
        a_legacy=LegacyDefinition(name='x_castep_core_spectra_n_bands'))

    x_castep_core_spectra_conv_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_castep_core_spectra_conv_tolerance
        ''',
        a_legacy=LegacyDefinition(name='x_castep_core_spectra_conv_tolerance'))


class x_castep_section_ts_scf_iteration(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_ts_scf_iteration'))

    x_castep_scf_ts_iteration_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy
        ''',
        a_legacy=LegacyDefinition(name='x_castep_scf_ts_iteration_energy'))

    x_castep_scf_ts_iteration_energy_change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy change
        ''',
        a_legacy=LegacyDefinition(name='x_castep_scf_ts_iteration_energy_change'))

    x_castep_scf_ts_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy time
        ''',
        a_legacy=LegacyDefinition(name='x_castep_scf_ts_time'))


class x_castep_section_tddft(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_tddft'))

    x_castep_tddft_iteration = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Iteration number
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_iteration'))

    x_castep_wall_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Iteration wall time (s)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_wall_time'))

    x_castep_state_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        state number
        ''',
        a_legacy=LegacyDefinition(name='x_castep_state_number'))

    x_castep_state_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        state energy
        ''',
        a_legacy=LegacyDefinition(name='x_castep_state_energy'))

    x_castep_state_energy_error = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        state energy error
        ''',
        a_legacy=LegacyDefinition(name='x_castep_state_energy_error'))

    x_castep_tddft_calculation_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        calculation time
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_calculation_time'))


class x_castep_section_atom_ionic_velocities(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_atom_ionic_velocities'))


class x_castep_section_atom_positions_optim(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_atom_positions_optim'))

    x_castep_cell_angle_alpha_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle alpha
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_angle_alpha_optim'))

    x_castep_cell_angle_beta_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle beta
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_angle_beta_optim'))

    x_castep_cell_angle_gamma_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle gamma
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_angle_gamma_optim'))

    x_castep_cell_length_a_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        a unit cell edge length
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_length_a_optim'))

    x_castep_cell_length_b_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        b unit cell edge length
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_length_b_optim'))

    x_castep_cell_length_c_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        c unit cell edge length
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_length_c_optim'))


class x_castep_section_atom_positions(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_atom_positions'))

    x_castep_cell_angle_alpha = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle alpha
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_angle_alpha'))

    x_castep_cell_angle_beta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle beta
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_angle_beta'))

    x_castep_cell_angle_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle gamma
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_angle_gamma'))

    x_castep_cell_length_a = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        a unit cell edge length
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_length_a'))

    x_castep_cell_length_b = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        b unit cell edge length
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_length_b'))

    x_castep_cell_length_c = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        c unit cell edge length
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_length_c'))


class x_castep_section_cell_optim(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_cell_optim'))

    x_castep_cell_vector_optim = Quantity(
        type=str,
        shape=[],
        description='''
        Temporay storage for cell vectors
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_vector_optim'))


class x_castep_section_cell(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_cell'))

    x_castep_cell_vector = Quantity(
        type=str,
        shape=[],
        description='''
        Temporay storage for cell vectors
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_vector'))


class x_castep_section_collect_scf_eigenvalues(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_collect_scf_eigenvalues'))


class x_castep_section_eigenvalues_1(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_eigenvalues_1'))

    x_castep_store_eigenvalues_1 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storing eigenvalues
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_eigenvalues_1'))


class x_castep_section_eigenvalues(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_eigenvalues'))

    x_castep_store_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storing eigenvalues
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_eigenvalues'))


class x_castep_section_functional_definition(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_functional_definition'))

    x_castep_functional_type = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional definition in CASTEP convention
        ''',
        a_legacy=LegacyDefinition(name='x_castep_functional_type'))

    x_castep_functional_weight = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional weight in CASTEP convention
        ''',
        a_legacy=LegacyDefinition(name='x_castep_functional_weight'))


class x_castep_section_functionals(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_functionals'))

    x_castep_functional_name = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional in CASTEP convention
        ''',
        a_legacy=LegacyDefinition(name='x_castep_functional_name'))


class x_castep_section_population_analysis(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_population_analysis'))

    x_castep_mulliken_atom_index = Quantity(
        type=str,
        shape=[],
        description='''
        Mulliken_atom_index
        ''',
        a_legacy=LegacyDefinition(name='x_castep_mulliken_atom_index'))

    x_castep_mulliken_atom = Quantity(
        type=str,
        shape=[],
        description='''
        Mulliken_atom kind
        ''',
        a_legacy=LegacyDefinition(name='x_castep_mulliken_atom'))

    x_castep_orbital_contributions = Quantity(
        type=str,
        shape=[],
        description='''
        Mulliken_contributions
        ''',
        a_legacy=LegacyDefinition(name='x_castep_orbital_contributions'))

    x_castep_orbital_s = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_contribution_orbital s
        ''',
        a_legacy=LegacyDefinition(name='x_castep_orbital_s'))

    x_castep_orbital_p = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_contribution_orbital p
        ''',
        a_legacy=LegacyDefinition(name='x_castep_orbital_p'))

    x_castep_orbital_d = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_contribution_orbital d
        ''',
        a_legacy=LegacyDefinition(name='x_castep_orbital_d'))

    x_castep_orbital_f = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_contribution_orbital f
        ''',
        a_legacy=LegacyDefinition(name='x_castep_orbital_f'))

    x_castep_total_orbital = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_total_contribution
        ''',
        a_legacy=LegacyDefinition(name='x_castep_total_orbital'))

    x_castep_mulliken_charge_store = Quantity(
        type=str,
        shape=[],
        description='''
        Mulliken_charges
        ''',
        a_legacy=LegacyDefinition(name='x_castep_mulliken_charge_store'))

    x_castep_mulliken_charge = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_charges
        ''',
        a_legacy=LegacyDefinition(name='x_castep_mulliken_charge'))


class x_castep_section_geom_optimisation_method(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_geom_optimisation_method'))

    x_castep_geometry_optim_method = Quantity(
        type=str,
        shape=[],
        description='''
        Determines optimisation method used
        ''',
        a_legacy=LegacyDefinition(name='x_castep_geometry_optim_method'))


class x_castep_section_optics_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_optics_parameters'))

    x_castep_optics_n_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        optics_number_of_bands
        ''',
        a_legacy=LegacyDefinition(name='x_castep_optics_n_bands'))

    x_castep_optics_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        optics_band_convergence_tolerance
        ''',
        a_legacy=LegacyDefinition(name='x_castep_optics_tolerance'))


class x_castep_section_tddft_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_tddft_parameters'))

    x_castep_tddft_n_excited_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of excited states
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_n_excited_states'))

    x_castep_tddft_n_states_forces = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of states for forces
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_n_states_forces'))

    x_castep_tddft_state_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        tolerance (eV)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_state_tolerance'))

    x_castep_tddft_state_tolerance_window = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        tolerance window iterations
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_state_tolerance_window'))

    x_castep_tddft_max_iter = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        max number iterations
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_max_iter'))

    x_castep_tddft_extra_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of extra states
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_extra_states'))

    x_castep_tddft_functional = Quantity(
        type=str,
        shape=[],
        description='''
        tddft functional
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_functional'))

    x_castep_tddft_method = Quantity(
        type=str,
        shape=[],
        description='''
        tddft method
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_method'))

    x_castep_tddft_eigenmethod = Quantity(
        type=str,
        shape=[],
        description='''
        tddft eigenmethod
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_eigenmethod'))

    x_castep_tddft_approximation = Quantity(
        type=str,
        shape=[],
        description='''
        tddft approximation
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_approximation'))

    x_castep_tddft_position_op = Quantity(
        type=str,
        shape=[],
        description='''
        tddft position operator
        ''',
        a_legacy=LegacyDefinition(name='x_castep_tddft_position_op'))


class x_castep_section_phonons(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_phonons'))

    x_castep_phonon_method = Quantity(
        type=str,
        shape=[],
        description='''
        Phonon calculation method
        ''',
        a_legacy=LegacyDefinition(name='x_castep_phonon_method'))

    x_castep_DFPT_solver_method = Quantity(
        type=str,
        shape=[],
        description='''
        Phonon DFPT solver method
        ''',
        a_legacy=LegacyDefinition(name='x_castep_DFPT_solver_method'))

    x_castep_phonon_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Phonon calculation tolerance (eV/A**2)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_phonon_tolerance'))

    x_castep_phonon_cycles = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Phonon calculation cycles
        ''',
        a_legacy=LegacyDefinition(name='x_castep_phonon_cycles'))

    x_castep_band_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Phonon band convergence tolerance window
        ''',
        a_legacy=LegacyDefinition(name='x_castep_band_tolerance'))


class x_castep_section_density_mixing_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_density_mixing_parameters'))

    x_castep_density_mixing_scheme = Quantity(
        type=str,
        shape=[],
        description='''
        density_mixing_scheme
        ''',
        a_legacy=LegacyDefinition(name='x_castep_density_mixing_scheme'))

    x_castep_density_mixing_length = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        density_mixing_scheme_length
        ''',
        a_legacy=LegacyDefinition(name='x_castep_density_mixing_length'))

    x_castep_charge_density_mixing_amplitude = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        charge_density_mixing_amplitude
        ''',
        a_legacy=LegacyDefinition(name='x_castep_charge_density_mixing_amplitude'))

    x_castep_cut_off_energy_for_mixing = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        charge_density_mixing_cut_off_energy_for_mixing (A)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cut_off_energy_for_mixing'))


class x_castep_section_population_analysis_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_population_analysis_parameters'))

    x_castep_population_analysis_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Population_analysis_cutoff_(A)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_population_analysis_cutoff'))


class x_castep_section_ts_scf(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_ts_scf'))

    x_castep_scf_ts_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy Total
        ''',
        a_legacy=LegacyDefinition(name='x_castep_scf_ts_total'))

    x_castep_scf_ts_total_energy_free = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy Total free
        ''',
        a_legacy=LegacyDefinition(name='x_castep_scf_ts_total_energy_free'))

    x_castep_scf_ts_T0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy T0 free
        ''',
        a_legacy=LegacyDefinition(name='x_castep_scf_ts_T0'))

    x_castep_section_ts_scf_iteration = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_scf_iteration'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_ts_scf_iteration'))


class x_castep_section_k_band(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_k_band'))


class x_castep_section_k_points_1(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_k_points_1'))

    x_castep_store_k_points_1 = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing k points coordinates (fractional)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_k_points_1'))


class x_castep_section_k_points(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_k_points'))

    x_castep_store_k_points = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing k points coordinates (fractional)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_k_points'))


class x_castep_section_relativity_treatment(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_relativity_treatment'))

    x_castep_relativity_treatment_scf = Quantity(
        type=str,
        shape=[],
        description='''
        Relativity treatment in CASTEP convention
        ''',
        a_legacy=LegacyDefinition(name='x_castep_relativity_treatment_scf'))


class x_castep_section_scf_eigenvalues(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_scf_eigenvalues'))

    x_castep_store_scf_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_scf_eigenvalues'))


class x_castep_section_SCF_iteration_frame(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_SCF_iteration_frame'))

    x_castep_frame_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_store_t_md_frame
        ''',
        a_legacy=LegacyDefinition(name='x_castep_frame_time'))

    x_castep_SCF_frame_energy = Quantity(
        type=str,
        shape=[],
        description='''
        energy_frame_iterations
        ''',
        a_legacy=LegacyDefinition(name='x_castep_SCF_frame_energy'))

    x_castep_SCF_frame_energy_gain = Quantity(
        type=str,
        shape=[],
        description='''
        energy_frame_iterations_gain
        ''',
        a_legacy=LegacyDefinition(name='x_castep_SCF_frame_energy_gain'))

    x_castep_frame_time_scf_iteration_wall_end = Quantity(
        type=str,
        shape=[],
        description='''
        energy_frame_wall_end_time
        ''',
        a_legacy=LegacyDefinition(name='x_castep_frame_time_scf_iteration_wall_end'))

    x_castep_frame_energy_free = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        energy_free
        ''',
        a_legacy=LegacyDefinition(name='x_castep_frame_energy_free'))

    x_castep_frame_energy_total_T0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        energy_free_corrected_for_finite_basis
        ''',
        a_legacy=LegacyDefinition(name='x_castep_frame_energy_total_T0'))


class x_castep_section_scf_k_points(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_scf_k_points'))

    x_castep_store_scf_k_points = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_scf_k_points'))

    x_castep_section_scf_eigenvalues = SubSection(
        sub_section=SectionProxy('x_castep_section_scf_eigenvalues'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_scf_eigenvalues'))


class x_castep_section_spin_number(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_spin_number'))

    x_castep_spin_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_castep_spin_number'))


class x_castep_section_stress_tensor(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_stress_tensor'))

    x_castep_store_stress_tensor = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing stress tensor components
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_stress_tensor'))


class x_castep_section_time(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_time'))

    x_castep_calculation_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        castep_calculation_time
        ''',
        a_legacy=LegacyDefinition(name='x_castep_calculation_time'))

    x_castep_finalisation_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        castep_finalisation_time
        ''',
        a_legacy=LegacyDefinition(name='x_castep_finalisation_time'))

    x_castep_initialisation_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_castep_Initialisation_time
        ''',
        a_legacy=LegacyDefinition(name='x_castep_initialisation_time'))


class x_castep_section_raman_tensor(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_raman_tensor'))

    x_castep_store_raman_tensor = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged Raman susceptibility tensor
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_raman_tensor'))

    x_castep_ramen_tensor = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='ampere / unified_atomic_mass_unit',
        description='''
        Ramen tensor scaled by 0.5
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ramen_tensor'))


class x_castep_section_scf_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_scf_parameters'))

    x_castep_energy_threshold_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Energy Threshold store
        ''',
        a_legacy=LegacyDefinition(name='x_castep_energy_threshold_store'))

    x_castep_max_iter_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of maximum iterations steps store
        ''',
        a_legacy=LegacyDefinition(name='x_castep_max_iter_store'))

    x_castep_smearing_kind = Quantity(
        type=str,
        shape=[],
        description='''
        Smearing kind
        ''',
        a_legacy=LegacyDefinition(name='x_castep_smearing_kind'))

    x_castep_smearing_width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Smearing width
        ''',
        a_legacy=LegacyDefinition(name='x_castep_smearing_width'))


class x_castep_section_electronic_spectroscpy_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_electronic_spectroscpy_parameters'))

    x_castep_theory_level = Quantity(
        type=str,
        shape=[],
        description='''
        Electronic spectroscopy parameters theory level
        ''',
        a_legacy=LegacyDefinition(name='x_castep_theory_level'))

    x_castep_spectroscopy_calculation = Quantity(
        type=str,
        shape=[],
        description='''
        Electronic spectroscopy parameters calculation type
        ''',
        a_legacy=LegacyDefinition(name='x_castep_spectroscopy_calculation'))

    x_castep_spec_max_iter = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Max number of iterations
        ''',
        a_legacy=LegacyDefinition(name='x_castep_spec_max_iter'))

    x_castep_spec_max_steps = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Max number of steps
        ''',
        a_legacy=LegacyDefinition(name='x_castep_spec_max_steps'))

    x_castep_spec_max_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Max number of bands
        ''',
        a_legacy=LegacyDefinition(name='x_castep_spec_max_bands'))

    x_castep_spec_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Tolerance (eV)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_spec_tolerance'))


class x_castep_section_md(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_md'))

    x_castep_md_energies = Quantity(
        type=str,
        shape=[],
        description='''
        md_energy_components
        ''',
        a_legacy=LegacyDefinition(name='x_castep_md_energies'))

    x_castep_md_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        md_temp(K)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_md_temperature'))

    x_castep_md_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        md_pressure
        ''',
        a_legacy=LegacyDefinition(name='x_castep_md_pressure'))

    x_castep_md_cell_vectors = Quantity(
        type=str,
        shape=[],
        description='''
        md_cell_vectors
        ''',
        a_legacy=LegacyDefinition(name='x_castep_md_cell_vectors'))

    x_castep_md_cell_vectors_vel = Quantity(
        type=str,
        shape=[],
        description='''
        md_cell_vectors_velocities
        ''',
        a_legacy=LegacyDefinition(name='x_castep_md_cell_vectors_vel'))

    x_castep_md_stress_tensor = Quantity(
        type=str,
        shape=[],
        description='''
        md_stress_tensor
        ''',
        a_legacy=LegacyDefinition(name='x_castep_md_stress_tensor'))

    x_castep_md_positions = Quantity(
        type=str,
        shape=[],
        description='''
        md_positions
        ''',
        a_legacy=LegacyDefinition(name='x_castep_md_positions'))

    x_castep_md_forces = Quantity(
        type=str,
        shape=[],
        description='''
        md_forces
        ''',
        a_legacy=LegacyDefinition(name='x_castep_md_forces'))

    x_castep_md_lab = Quantity(
        type=str,
        shape=[],
        description='''
        md_lables
        ''',
        a_legacy=LegacyDefinition(name='x_castep_md_lab'))

    x_castep_md_veloc = Quantity(
        type=str,
        shape=[],
        description='''
        md_veloc
        ''',
        a_legacy=LegacyDefinition(name='x_castep_md_veloc'))


class x_castep_section_ts(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_ts'))

    x_castep_ts_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_total
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_energy_total'))

    x_castep_ts_cell_vectors = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        ts_cell_vectors
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_cell_vectors'))

    x_castep_ts_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_positions
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_positions'))

    x_castep_ts_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_forces
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_forces'))

    x_castep_ts_lab = Quantity(
        type=str,
        shape=[],
        description='''
        ts_lables
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_lab'))


class x_castep_section_ts_store(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_ts_store'))

    x_castep_ts_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_ts_store
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_energy'))

    x_castep_ts_cell_vectors_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_cell_vectors
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_cell_vectors_store'))

    x_castep_ts_positions_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_positions_store
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_positions_store'))

    x_castep_ts_forces_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_forces_store
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_forces_store'))


class x_castep_section_ts_final_store(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_ts_final_store'))

    x_castep_ts_energy_final_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_ts
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_energy_final_store'))

    x_castep_ts_forces_final_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_forces_final
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_forces_final_store'))

    x_castep_ts_positions_final_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_energy_positions_final
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_positions_final_store'))

    x_castep_ts_cell_vectors_final_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_cell_vectors_final
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_cell_vectors_final_store'))


class x_castep_section_ts_final(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_ts_final'))

    x_castep_ts_energy_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_final
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_energy_final'))

    x_castep_ts_cell_vectors_final = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        ts_cell_vectors
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_cell_vectors_final'))

    x_castep_ts_positions_final = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_positions_final
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_positions_final'))

    x_castep_ts_forces_final = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_force_finals
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_forces_final'))


class x_castep_section_ts_product(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_ts_product'))

    x_castep_ts_energy_product = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_prod
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_energy_product'))

    x_castep_ts_forces_product = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_energy_forces_pro
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_forces_product'))

    x_castep_ts_positions_product = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_energy_positions_pro
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_positions_product'))

    x_castep_ts_cell_vectors_product = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        ts_product_cell_vectors
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_cell_vectors_product'))


class x_castep_section_ts_product_store(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_ts_product_store'))

    x_castep_ts_energy_product_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_pro
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_energy_product_store'))

    x_castep_ts_forces_pro_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_energy_forces_pro_store
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_forces_pro_store'))

    x_castep_ts_positions_pro_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_energy_positions_pro_store
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_positions_pro_store'))

    x_castep_ts_cell_vectors_pro_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_cell_vectors_pro
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_cell_vectors_pro_store'))


class x_castep_section_ts_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_ts_parameters'))

    x_castep_ts_method = Quantity(
        type=str,
        shape=[],
        description='''
        ts_method
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_method'))

    x_castep_ts_protocol = Quantity(
        type=str,
        shape=[],
        description='''
        ts_protocol
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_protocol'))

    x_castep_ts_number_qst = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_qst_iterations
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_number_qst'))

    x_castep_ts_number_cg = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_number_of_cg_iterations
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_number_cg'))

    x_castep_ts_force_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_force_tolerance (eV/A)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_force_tolerance'))

    x_castep_ts_displacement_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_displacement_tolerance (A)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_displacement_tolerance'))


class x_castep_section_DFT_SEDC(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_DFT_SEDC'))

    x_castep_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        correlation energy
        ''',
        a_legacy=LegacyDefinition(name='x_castep_correction_energy'))

    x_castep_de_atom = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dE/atom
        ''',
        a_legacy=LegacyDefinition(name='x_castep_de_atom'))

    x_castep_dfmax_atom = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dfmax/atom
        ''',
        a_legacy=LegacyDefinition(name='x_castep_dfmax_atom'))

    x_castep_structure_energy_corr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        structure energy correction
        ''',
        a_legacy=LegacyDefinition(name='x_castep_structure_energy_corr'))

    x_castep_PBC_image_inter_corr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        PBC image interaction corr.
        ''',
        a_legacy=LegacyDefinition(name='x_castep_PBC_image_inter_corr'))

    x_castep_total_energy_correction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total energy correction
        ''',
        a_legacy=LegacyDefinition(name='x_castep_total_energy_correction'))

    x_castep_total_fmax_correction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        correction F max ev/A
        ''',
        a_legacy=LegacyDefinition(name='x_castep_total_fmax_correction'))

    x_castep_shell = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        shell
        ''',
        a_legacy=LegacyDefinition(name='x_castep_shell'))

    x_castep_total_dispersion_corrected_free_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total_dispersion_corrected_free_energy
        ''',
        a_legacy=LegacyDefinition(name='x_castep_total_dispersion_corrected_free_energy'))

    x_castep_disp_corrected_energy_total_T0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dispersion corrected zero point
        ''',
        a_legacy=LegacyDefinition(name='x_castep_disp_corrected_energy_total_T0'))


class x_castep_section_van_der_Waals_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_castep_section_van_der_Waals_parameters'))

    x_castep_disp_method_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name type
        ''',
        a_legacy=LegacyDefinition(name='x_castep_disp_method_name'))

    x_castep_Parameter_d = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method G06
        ''',
        a_legacy=LegacyDefinition(name='x_castep_Parameter_d'))

    x_castep_Parameter_LAMBDA = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method OBS
        ''',
        a_legacy=LegacyDefinition(name='x_castep_Parameter_LAMBDA'))

    x_castep_Parameter_n = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method OBS
        ''',
        a_legacy=LegacyDefinition(name='x_castep_Parameter_n'))

    x_castep_Parameter_s6 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method G06
        ''',
        a_legacy=LegacyDefinition(name='x_castep_Parameter_s6'))

    x_castep_Parameter_sR = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method TS
        ''',
        a_legacy=LegacyDefinition(name='x_castep_Parameter_sR'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_castep_atom_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Forces on the atoms as minus gradient of energy_total, including forces' unitary-
        transformation (rigid body) filtering and including constraints, if present. The
        derivatives with respect to displacements of the nuclei in the gradient are
        evaluated according to the coordinate system defined in coordinate_system. In
        addition, these forces are obtained by filtering out the unitary transformations
        (translations of the center of mass and rigid rotations of the whole system, when
        non periodic), atom_forces_raw for the unfiltered counterpart. Furthermore, forces
        due to constraints like fixed atoms, distances, angles, dihedrals, and so on, are
        here included (see atom_forces_raw for the unfiltered counterpart).
        ''',
        a_legacy=LegacyDefinition(name='x_castep_atom_forces'))

    x_castep_basis_set_planewave_cutoff_iteration_0 = Quantity(
        type=str,
        shape=[],
        description='''
        cutoff at iteration 0 of geometry optimisation
        ''',
        a_legacy=LegacyDefinition(name='x_castep_basis_set_planewave_cutoff_iteration_0'))

    x_castep_compiler = Quantity(
        type=str,
        shape=[],
        description='''
        Compiler name
        ''',
        a_legacy=LegacyDefinition(name='x_castep_compiler'))

    x_castep_constants_reference = Quantity(
        type=str,
        shape=[],
        description='''
        Fundamental constant data source
        ''',
        a_legacy=LegacyDefinition(name='x_castep_constants_reference'))

    x_castep_fft_library = Quantity(
        type=str,
        shape=[],
        description='''
        fft library name
        ''',
        a_legacy=LegacyDefinition(name='x_castep_fft_library'))

    x_castep_geom_converged = Quantity(
        type=str,
        shape=[],
        description='''
        CASTEP_geom_converged
        ''',
        a_legacy=LegacyDefinition(name='x_castep_geom_converged'))

    x_castep_maths_library = Quantity(
        type=str,
        shape=[],
        description='''
        Maths library name
        ''',
        a_legacy=LegacyDefinition(name='x_castep_maths_library'))

    x_castep_program_compilation_date = Quantity(
        type=str,
        shape=[],
        description='''
        Compilation date (string)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_program_compilation_date'))

    x_castep_program_compilation_time = Quantity(
        type=str,
        shape=[],
        description='''
        Compilation time (string)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_program_compilation_time'))

    x_castep_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        Run execution date (string)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_program_execution_date'))

    x_castep_program_execution_time = Quantity(
        type=str,
        shape=[],
        description='''
        Run execution start time (string)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_program_execution_time'))

    x_castep_crystal_point_group = Quantity(
        type=str,
        shape=[],
        description='''
        Point group of the crystal (Schoenflies notation)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_crystal_point_group'))

    x_castep_space_group = Quantity(
        type=str,
        shape=[],
        description='''
        Point space of the crystal
        ''',
        a_legacy=LegacyDefinition(name='x_castep_space_group'))

    x_castep_ts_path = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_path_number
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_path'))

    x_castep_ts_path_ts_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_path_final
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_path_ts_final'))

    x_castep_ts_path_product = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_path_pro
        ''',
        a_legacy=LegacyDefinition(name='x_castep_ts_path_product'))

    x_castep_store_atom_forces_band = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_atom_forces_band'))

    x_castep_elec_methd = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing electronic structure method
        ''',
        a_legacy=LegacyDefinition(name='x_castep_elec_methd'))

    x_castep_section_vibrational_frequencies = SubSection(
        sub_section=SectionProxy('x_castep_section_vibrational_frequencies'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_vibrational_frequencies'))

    x_castep_section_band_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_band_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_band_parameters'))

    x_castep_section_core_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_core_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_core_parameters'))

    x_castep_section_collect_scf_eigenvalues = SubSection(
        sub_section=SectionProxy('x_castep_section_collect_scf_eigenvalues'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_collect_scf_eigenvalues'))

    x_castep_section_population_analysis = SubSection(
        sub_section=SectionProxy('x_castep_section_population_analysis'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_population_analysis'))

    x_castep_section_geom_optimisation_method = SubSection(
        sub_section=SectionProxy('x_castep_section_geom_optimisation_method'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_geom_optimisation_method'))

    x_castep_section_optics_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_optics_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_optics_parameters'))

    x_castep_section_tddft_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_tddft_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_tddft_parameters'))

    x_castep_section_phonons = SubSection(
        sub_section=SectionProxy('x_castep_section_phonons'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_phonons'))

    x_castep_section_density_mixing_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_density_mixing_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_density_mixing_parameters'))

    x_castep_section_population_analysis_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_population_analysis_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_population_analysis_parameters'))

    x_castep_section_SCF_iteration_frame = SubSection(
        sub_section=SectionProxy('x_castep_section_SCF_iteration_frame'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_SCF_iteration_frame'))

    x_castep_section_time = SubSection(
        sub_section=SectionProxy('x_castep_section_time'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_time'))

    x_castep_section_raman_tensor = SubSection(
        sub_section=SectionProxy('x_castep_section_raman_tensor'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_raman_tensor'))

    x_castep_section_scf_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_scf_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_scf_parameters'))

    x_castep_section_electronic_spectroscpy_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_electronic_spectroscpy_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_electronic_spectroscpy_parameters'))

    x_castep_section_md = SubSection(
        sub_section=SectionProxy('x_castep_section_md'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_md'))

    x_castep_section_ts = SubSection(
        sub_section=SectionProxy('x_castep_section_ts'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_ts'))

    x_castep_section_ts_store = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_store'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_ts_store'))

    x_castep_section_ts_final_store = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_final_store'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_ts_final_store'))

    x_castep_section_ts_final = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_final'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_ts_final'))

    x_castep_section_ts_product = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_product'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_ts_product'))

    x_castep_section_ts_product_store = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_product_store'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_ts_product_store'))

    x_castep_section_ts_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_ts_parameters'))

    x_castep_section_van_der_Waals_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_van_der_Waals_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_van_der_Waals_parameters'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_castep_net_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Net charge of system
        ''',
        a_legacy=LegacyDefinition(name='x_castep_net_charge'))

    x_castep_number_of_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of bands
        ''',
        a_legacy=LegacyDefinition(name='x_castep_number_of_bands'))

    x_castep_number_of_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of electrons
        ''',
        a_legacy=LegacyDefinition(name='x_castep_number_of_electrons'))

    x_castep_atom_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Storing atomic positions in fractional coordinates
        ''',
        a_legacy=LegacyDefinition(name='x_castep_atom_positions'))

    x_castep_cell_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_cell_volume
        ''',
        a_legacy=LegacyDefinition(name='x_castep_cell_volume'))

    x_castep_optimised_atom_labels = Quantity(
        type=str,
        shape=['number_of_atoms'],
        description='''
        Temporary storing atomic positions
        ''',
        a_legacy=LegacyDefinition(name='x_castep_optimised_atom_labels'))

    x_castep_optimised_atom_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Storing atomic optimised positions in fractional coordinates
        ''',
        a_legacy=LegacyDefinition(name='x_castep_optimised_atom_positions'))

    x_castep_velocities_cell_vector = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        cell vector velocities
        ''',
        a_legacy=LegacyDefinition(name='x_castep_velocities_cell_vector'))

    x_castep_store_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atom labels
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_atom_labels'))

    x_castep_store_atom_number = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atom labels
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_atom_number'))

    x_castep_store_atom_positions = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_atom_positions'))

    x_castep_store_atom_ionic_velocities = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_atom_ionic_velocities'))

    x_castep_atom_ionic_velocities = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Temporary storing atomic positions
        ''',
        a_legacy=LegacyDefinition(name='x_castep_atom_ionic_velocities'))

    x_castep_store_optimised_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_optimised_atom_labels'))

    x_castep_store_optimised_atom_positions = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_optimised_atom_positions'))

    x_castep_section_tddft = SubSection(
        sub_section=SectionProxy('x_castep_section_tddft'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_tddft'))

    x_castep_section_atom_ionic_velocities = SubSection(
        sub_section=SectionProxy('x_castep_section_atom_ionic_velocities'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_atom_ionic_velocities'))

    x_castep_section_atom_positions_optim = SubSection(
        sub_section=SectionProxy('x_castep_section_atom_positions_optim'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_atom_positions_optim'))

    x_castep_section_atom_positions = SubSection(
        sub_section=SectionProxy('x_castep_section_atom_positions'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_atom_positions'))

    x_castep_section_cell_optim = SubSection(
        sub_section=SectionProxy('x_castep_section_cell_optim'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_cell_optim'))

    x_castep_section_cell = SubSection(
        sub_section=SectionProxy('x_castep_section_cell'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_cell'))


class section_basis_set_cell_dependent(public.section_basis_set_cell_dependent):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_basis_set_cell_dependent'))

    x_castep_basis_set_planewave_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing plane wave cutoff as string
        ''',
        a_legacy=LegacyDefinition(name='x_castep_basis_set_planewave_cutoff'))

    x_castep_size_std_grid = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        size of standard grid (eV)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_size_std_grid'))

    x_castep_size_fine_grid = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        sise of fine grid (1/A)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_size_fine_grid'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_castep_enthalpy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_Enthalpy
        ''',
        a_legacy=LegacyDefinition(name='x_castep_enthalpy'))

    x_castep_frequency = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_frequency (cm-1)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_frequency'))

    x_castep_improved_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        md_forces
        ''',
        a_legacy=LegacyDefinition(name='x_castep_improved_energy_total'))

    x_castep_frame_time_0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_store_t_md_frame
        ''',
        a_legacy=LegacyDefinition(name='x_castep_frame_time_0'))

    x_castep_geom_iteration_index = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Index for number of iterations in geometry optimisation
        ''',
        a_legacy=LegacyDefinition(name='x_castep_geom_iteration_index'))

    x_castep_store_atom_forces = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_atom_forces'))

    x_castep_initial_scf_iteration_wall_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Initial SCF iteration wall time
        ''',
        a_legacy=LegacyDefinition(name='x_castep_initial_scf_iteration_wall_time'))

    x_castep_total_dispersion_corrected_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total electronic energy that includes dispersion energy computed with
        Disp_method_name not corrected for finite basis-set
        ''',
        a_legacy=LegacyDefinition(name='x_castep_total_dispersion_corrected_energy'))

    x_castep_total_energy_corrected_for_finite_basis = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_total_energy_corrected_for_finite_basis
        ''',
        a_legacy=LegacyDefinition(name='x_castep_total_energy_corrected_for_finite_basis'))

    x_castep_total_energy_corrected_for_finite_basis_store = Quantity(
        type=str,
        shape=[],
        description='''
        CASTEP_total_energy_corrected_for_finite_basis_store
        ''',
        a_legacy=LegacyDefinition(name='x_castep_total_energy_corrected_for_finite_basis_store'))

    energy_total_scf_iteration_list = Quantity(
        type=np.dtype(np.float64),
        shape=[-1],
        description='''
        Total electronic energy calculated with XC_method_scf during the scf iterations is
        stored in a list
        ''',
        a_legacy=LegacyDefinition(name='energy_total_scf_iteration_list'))

    x_castep_section_ts_scf = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_scf'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_ts_scf'))

    x_castep_section_scf_k_points = SubSection(
        sub_section=SectionProxy('x_castep_section_scf_k_points'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_scf_k_points'))

    x_castep_section_spin_number = SubSection(
        sub_section=SectionProxy('x_castep_section_spin_number'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_spin_number'))

    x_castep_section_stress_tensor = SubSection(
        sub_section=SectionProxy('x_castep_section_stress_tensor'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_stress_tensor'))

    x_castep_section_DFT_SEDC = SubSection(
        sub_section=SectionProxy('x_castep_section_DFT_SEDC'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_DFT_SEDC'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_castep_functional_and_weight = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional+weight in CASTEP convention
        ''',
        a_legacy=LegacyDefinition(name='x_castep_functional_and_weight'))

    x_castep_section_functional_definition = SubSection(
        sub_section=SectionProxy('x_castep_section_functional_definition'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_functional_definition'))

    x_castep_section_functionals = SubSection(
        sub_section=SectionProxy('x_castep_section_functionals'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_functionals'))

    x_castep_section_relativity_treatment = SubSection(
        sub_section=SectionProxy('x_castep_section_relativity_treatment'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_relativity_treatment'))


class section_sampling_method(public.section_sampling_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_sampling_method'))

    x_castep_geometry_stress_com_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        tolerance for stress components in geometry optimisation (GPa)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_geometry_stress_com_tolerance'))

    x_castep_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        thermostat_target_temperature(K)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_thermostat_target_temperature'))

    x_castep_barostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        barostat_type
        ''',
        a_legacy=LegacyDefinition(name='x_castep_barostat_type'))

    x_castep_thermostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        thermostat_type
        ''',
        a_legacy=LegacyDefinition(name='x_castep_thermostat_type'))

    x_castep_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        thermostat_type
        ''',
        a_legacy=LegacyDefinition(name='x_castep_thermostat_tau'))

    x_castep_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        barostat_tau
        ''',
        a_legacy=LegacyDefinition(name='x_castep_barostat_tau'))

    x_castep_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        MD_time_step (ps)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_integrator_dt'))

    x_castep_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        MD_time_step_number
        ''',
        a_legacy=LegacyDefinition(name='x_castep_number_of_steps_requested'))

    x_castep_max_number_of_steps = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number_of iterations geom_optim
        ''',
        a_legacy=LegacyDefinition(name='x_castep_max_number_of_steps'))

    x_castep_frame_pressure = Quantity(
        type=str,
        shape=[],
        description='''
        MD_pressure
        ''',
        a_legacy=LegacyDefinition(name='x_castep_frame_pressure'))

    x_castep_frame_energy_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        MD_scf_energy tolerance (eV)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_frame_energy_tolerance'))

    x_castep_frame_eigen_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        MD_scf_eigen tolerance (eV)
        ''',
        a_legacy=LegacyDefinition(name='x_castep_frame_eigen_tolerance'))


class section_k_band(public.section_k_band):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_k_band'))

    x_castep_k_path = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_castep_k_path'))

    x_castep_store_k_path = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_k_path'))

    x_castep_store_k_label = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_k_label'))

    x_castep_section_eigenvalues_1 = SubSection(
        sub_section=SectionProxy('x_castep_section_eigenvalues_1'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_eigenvalues_1'))

    x_castep_section_eigenvalues = SubSection(
        sub_section=SectionProxy('x_castep_section_eigenvalues'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_eigenvalues'))

    x_castep_section_k_band = SubSection(
        sub_section=SectionProxy('x_castep_section_k_band'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_k_band'))

    x_castep_section_k_points_1 = SubSection(
        sub_section=SectionProxy('x_castep_section_k_points_1'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_k_points_1'))

    x_castep_section_k_points = SubSection(
        sub_section=SectionProxy('x_castep_section_k_points'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_castep_section_k_points'))


class section_topology(common.section_topology):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_topology'))

    x_castep_store_atom_mass = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atom mass
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_atom_mass'))

    x_castep_store_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atom name
        ''',
        a_legacy=LegacyDefinition(name='x_castep_store_atom_name'))


m_package.__init_metainfo__()



# Copyright 2015-2018 Mikkel Strange, Fawzi Mohamed, Ankit Kariryaa, Ask Hjorth Larsen, Jens Jørgen Mortensen
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

f2p_version = {6: '1.1.0',
               5: '0.11.0',
               3: '0.10.0'}


def get_prog_version(version):
    if isinstance(version, int):
        return f2p_version[version]
    else:
        return '0.9.0'


if __name__ == '__main__':
    print(get_prog_version(6))
    print(get_prog_version(0.3))



from builtins import object
from nomadcore.simple_parser import mainFunction, CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
import os, sys, json


################################################################
# This is the subparser for the WIEN2k input file (.in1)
################################################################

# Copyright 2016-2018 Daria M. Tomecka, Fawzi Mohamed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Daria M. Tomecka"
__maintainer__ = "Daria M. Tomecka"
__email__ = "tomeckadm@gmail.com;"
__date__ = "15/05/2017"

class Wien2kIn1Context(object):
    """context for wien2k In1 parser"""

    def __init__(self):
        self.parser = None

    def initialize_values(self):
        """allows to reset values if the same superContext is used to parse different files"""
        pass

    def startedParsing(self, path, parser):
        """called when parsing starts"""
        self.parser = parser
        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()


# description of the input
def buildIn1Matchers():
    return SM(
    name = 'root',
    weak = True,
    startReStr = "",
        sections = ["section_run", "section_method"],
    subMatchers = [
#        SM(name = 'systemName',
#          startReStr = r"(?P<x_wien2k_system_nameIn>.*)"),
        SM(r"\s*(?P<x_wien2k_wf_switch>[\w*]+)\s*EF=[-+.0-9]*\s*\W*WFFIL, WFPRI, ENFIL, SUPWF\W"),
        SM(r"\s*(?P<x_wien2k_rkmax>[0-9.]+)\s*[0-9]+\s*[0-9]+\s*\WR-..\WK-...; MAX \w*.*")
#        SM(r"(?P<x_wien2k_calc_mode>.*)"),
#        SM(r"\s*(?P<x_wien2k_unit_cell_param_a>[-+0-9]*\.\d{0,6}){0,10}\s*(?P<x_wien2k_unit_cell_param_b>[-+0-9]*\.\d{0,6}){0,10}\s*(?P<x_wien2k_unit_cell_param_c>[-+0-9]*\.\d{0,6}){0,10}\s*(?P<x_wien2k_angle_between_unit_axis_alfa>[-+]?[0-9]*\.\d{0,6}){0,10}\s*(?P<x_wien2k_angle_between_unit_axis_beta>[-+]?[0-9]*\.\d{0,6}){0,10}\s*(?P<x_wien2k_angle_between_unit_axis_gamma>[-+]?[0-9]*\.\d*)"),
#        SM(r"\s*ATOM\s*[-0-9]+:\s*X=(?P<x_wien2k_atom_pos_x>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*Y=(?P<x_wien2k_atom_pos_y>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*Z=(?P<x_wien2k_atom_pos_z>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)",
#           repeats=True,
#           sections=["x_wien2k_section_equiv_atoms"],
#           subMatchers=[
#               SM(r"\s*[-0-9]+:\s*X=(?P<x_wien2k_atom_pos_x>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*Y=(?P<x_wien2k_atom_pos_y>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*Z=(?P<x_wien2k_atom_pos_z>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)",
#                  repeats=True
#              ),
#               SM(r"\s*(?P<x_wien2k_atom_name>^.+)\s*NPT=\s*(?P<x_wien2k_NPT>[0-9]+)\s*R0=(?P<x_wien2k_R0>[0-9.]+)\s*RMT=\s*(?P<x_wien2k_RMT>[0-9.]+)\s*Z:\s*(?P<x_wien2k_atomic_number_Z>[0-9.]+)",)
#           ]
#       )
    ])

def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                               'section_run': CachingLvl,
                               'section_method': CachingLvl
                              }
    return cachingLevelForMetaName

# loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json



import os, logging

from nomadcore.local_meta_info import InfoKindEl, InfoKindEnv, loadJsonFile

logger = logging.getLogger(__name__)
baseDir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
metaInfoDir = os.path.normpath(os.path.join(baseDir, "../nomad_meta_info"))

allMetaInfo = {}
for dirpath, dirnames, filenames in os.walk(metaInfoDir):
    for filename in filenames:
        if not filename.endswith(".nomadmetainfo.json"):
            continue
        filepath = os.path.join(dirpath, filename)
        try:
            newEnv, warnings = loadJsonFile(filepath, dependencyLoader = None, extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS, uri = None)
            if warnings.get("hasWarnings", True):
                logger.warn("loading %s had the following warnings: %s" %
                            (filepath, warnings))
            allMetaInfo[filepath] = newEnv
        except:
            logger.exception("nomadcore.basic_meta_info could not load %s", filepath)
baseMetaInfo = allMetaInfo[os.path.join(metaInfoDir, "nomad_base.nomadmetainfo.json")]
metaMetaInfo = allMetaInfo[os.path.join(metaInfoDir, "meta_types.nomadmetainfo.json")]



from typing import Dict
from warnings import warn

from optimade.models import Species as OptimadeStructureSpecies
from optimade.models import StructureResource as OptimadeStructure

from optimade.adapters.exceptions import ConversionError

try:
    from ase import Atoms, Atom
except (ImportError, ModuleNotFoundError):
    Atoms = None
    ASE_NOT_FOUND = "ASE not found, cannot convert structure to an ASE Atoms"


__all__ = ("get_ase_atoms",)


def get_ase_atoms(optimade_structure: OptimadeStructure) -> Atoms:
    """ Get ASE Atoms from OPTIMADE structure

    NOTE: Cannot handle partial occupancies (this includes vacancies)

    :param optimade_structure: OPTIMADE structure
    :return: ASE.Atoms
    """
    if globals().get("Atoms", None) is None:
        warn(ASE_NOT_FOUND)
        return None

    attributes = optimade_structure.attributes

    # Cannot handle partial occupancies
    if "disorder" in attributes.structure_features:
        raise ConversionError(
            "ASE cannot handle structures with partial occupancies, sorry."
        )

    species: Dict[str, OptimadeStructureSpecies] = {
        species.name: species for species in attributes.species
    }

    # Since we've made sure there are no species with more than 1 chemical symbol,
    # asking for index 0 will always work.
    if "X" in [specie.chemical_symbols[0] for specie in species.values()]:
        raise ConversionError(
            "ASE cannot handle structures with unknown ('X') chemical symbols, sorry."
        )

    atoms = []
    for site_number in range(attributes.nsites):
        species_name = attributes.species_at_sites[site_number]
        site = attributes.cartesian_site_positions[site_number]

        current_species = species[species_name]

        atoms.append(
            Atom(symbol=species_name, position=site, mass=current_species.mass)
        )

    return Atoms(
        symbols=atoms, cell=attributes.lattice_vectors, pbc=attributes.dimension_types
    )
